#!/p/perl/perl -w
#
# ======================================================================
# This file is copyright 1998 by the Purdue Research Foundation and may
# only be used under license.  For terms of the license, see the file
# named COPYRIGHT included with this software release.
# ======================================================================
#
# AAFID::Entity package.
#
# AAFID project, COAST Laboratory, 1998.
# 
# Diego Zamboni, Jan 26, 1998.
#
# $Id: Entity.pm,v 1.36 1998/09/28 15:39:49 zamboni Exp $
#
# NOTE: This file is in Perl's POD format. For more information, see the
#       perlpod(1) man page.
#

package AAFID::Entity;

# The following keeps up with the RCS version number. The self-assignment
# keeps the -w switch from complaining (because $VERSION may not be used
# here, but it is used in our base class).
$VERSION = do { my @r = (q$Revision: 1.36 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; $VERSION = $VERSION;

use vars qw(
	    @ISA
	    @EXPORT
	    %PARAMETERS
	    $VERSION
	    $sleeping
	    %_visitedClassPARAMETERS
	   );

use strict;
use Carp;
use AAFID::Common;
use AAFID::Comm;
use AAFID::Message;
use Sys::Hostname;
use Data::Dumper;
use Exporter();
use AAFID::Log;
use IO::File;
use IO::Socket;
use subs qw(init
	    sendConnectMessage
	    sendDisconnectMessage
	    ID
	    run
	    Cleanup
	    processInput
	    sendReport
	    setParameters
	    getParameter
	    getParametersList
	    sendMsgTo
	    );

@ISA=qw(
	Exporter
	AAFID::Log
	AAFID::Config
       );

@EXPORT=qw(_EndOfEntity);

=pod

This is the base class for all the AAFID entities: agents, transceivers
and monitors. All the classes that implement those entities should
be derived from this, because it provides all the common functionality.

=head1 Object creation

As is common practice in Perl, the object itself is represented by an
anonymous hash, which contains all the data pertaining to the object.

The B<AAFID::Entity> class automatically keeps a count of the number
of objects of each entity type (this is, B<Entity> or any derived
class), and assigns a number to each created entity. Also, a
(hopefully) unique identifier is generated for each entity with the
form C<hostname:ClassID:ClassVersion:InstanceNumber>. These are stored
in the object itself in a number of predefined parameters. See the
section "Standard Entity Parameters" for a full description of them.

The constructor is in charge of setting these parameters.

The constructor also sets up a hash called C<%Params> for access to
the parameters by calling the C<setTie> method. See the section
"Parameter manipulation" for details.

The constructor also calls the C<Init> method of the entity. This method
should be overriden by each specific class to provide entity-specific
initialization.

Every entity, when created, has to send a CONNECT message "up", this
is, to the entity that controls it, to let it know that it has
completed its initialization. This is not done at time of
construction, but instead in the C<run> method, just before starting
to run.

B<Important:> subclasses of B<AAFID::Entity> should not override the
constructor.

=cut

%PARAMETERS=(
	     Description	=>	"Generic Entity object",
	    );

sub new {
    my $class=shift;
    my $self={};
    my $countername="${class}::_EntityCounter";
    my $number;
    my $numberref;
    {
      no strict 'refs';
      $number=$ {$countername}++;
      $numberref=\$ {$countername};
    }
    $Data::Dumper::Indent=0;
    $Data::Dumper::Terse=1;
    bless $self, $class;
    # Incorporate the default %PARAMETERS hash into the entity
    # parameters.  We do this recursively, by following the
    # inheritance chain from top to bottom. In this way, the hash
    # defined by a subclass overrides the hash defined by a
    # superclass. Also, we add the ClassVersion parameter at the end,
    # because we want to take it from the $VERSION variable,
    # overriding anything else.
    {
      %_visitedClassPARAMETERS=();
      $self->_takePARAMETERS($class);
      no strict 'refs';
      my $vername="${class}::VERSION";
      $self->setParameters(ClassVersion => ($$vername || "0.00"));
    }
    # Now set the parameters specific to this entity.
    $self->setParameters(InstanceNumber => $number,
			 ClassCount     => $numberref,
			 ClassID	=> $class,
			 OpenHandles	=> IO::Select->new(),
			 OpenHandleIDs	=> {},
			 );
    # EntityID is set in a separate call because $self->ID needs
    # the other parameters to be in place.
    $self->setParameters(EntityID => $self->ID);
    # Load configuration information
    $self->configure();
    # Register and activate log categories.
    my $logfile=$self->getParameter('LogFile');
    $self->Log_register("signals", $logfile);
    $self->Log_register("I/O", $logfile);
    $self->Log_register("messages", $logfile);
    $self->Log_register("errors", $logfile);
    $self->Log_register("debug", $logfile);
    $self->Log_register("params", $logfile);
    $self->Log_register("processes", $logfile);
    # Activate some logging categories
    foreach (@{$self->getParameter('LogCategories')}) {
      $self->Log_activate($_);
    }
    push @INC, $self->getParametersList('ClassDir', 'AgentsDir');
    # Create class hash %Params tied to the parameters.
    $self->setTie();
    # Do local initialization.
    if ($self->_init) {
       return $self;
    } else {
       return undef;
    } 
}

=pod

To incorporate all the C<%PARAMETERS> hashes of the subclasses down to
the current class, we visit the graph defined by the C<@ISA> array
in a depth-first-left-to-right fashion (where left-to-right refers to
the order in which the elements of each C<@ISA> array are visited).
A cache is kept of classes that have been already visited in order
to avoid revisiting them in case they appear more than once in 
the inheritance graph.

=cut

sub _takePARAMETERS {
  my $self=shift;
  my $classname=shift;
  my $class;
  no strict 'refs';
  # First visit all the superclasses.
  foreach $class (@{"${classname}::ISA"}) {
    $self->_takePARAMETERS($class);
  }
  # And now take my own %PARAMETERS hash.
  if (!exists($_visitedClassPARAMETERS{$classname})) {
    $self->setParameters(%{"${classname}::PARAMETERS"});
    $_visitedClassPARAMETERS{$classname}=1;
  }
}

=pod

The B<AAFID::Entity> class provides an empty C<Init> method that functions
only as a place holder.

=cut

sub Init {
    # Put your own initialization code here.
}

sub _init {
  my $self=shift;
  ref($self) or croak "AAFID::_init can only be called as an instance method";
  $self->Init;
}

=head1 Sending a CONNECT message

The CONNECT message is a standard part of the AAFID communication protocol,
and as such the functionality of sending it is automatic, and the format
of the message is determined by the B<Entity> class. The format of the
DATA field for a CONNECT message is as follows:

    ClassID Description

The C<sendConnectMessage> subroutine sends one such message. If no
arguments are given, the message is sent up, and the message subtype
is "CHILD". If an argument is given, it is interpreted as an entity ID
if it is a string, and the message is sent to that entity, if its
handle is registered in our C<OpenHandleIDs> parameter, with a subtype
of "PARENT". If an argument is given but it is a reference, it is considered
to be a file handle reference, and the message is sent directly to that
handle.

=cut

sub sendConnectMessage {
    my $self=checkref(shift);
    my $to=shift;
    my $data=$self->getParameter("ClassID") . " " .
             $self->getParameter("Description");
    my $msg=$self->newMsg(TYPE		=> "CONNECT",
			  SUBTYPE	=> ($to?"PARENT":"CHILD"),
			  DATA		=> $data);
    if ($to) {
      if (!ref($to)) {
	# It is a scalar, use it as ID.
	$self->Log("I/O", 
		   "Sending CONNECT message to $to: ".$msg->toString."\n");
	$self->sendMsgTo($msg, $to)
      }
      else {
	# It is a reference, use it as file handle.
	$self->Log("I/O",
		   "Sending CONNECT message to handle " . Dumper($to) . ": ".
		   $msg->toString . "\n");
	AAFID::Comm::sendMsgTo($msg, $to);
      }
    }
    else {
      $self->Log("I/O", "Sending CONNECT message up: ".$msg->toString."\n");
      $self->sendReport($msg);
    }
    return $self;
}

=pod

The C<newMsg> subroutine returns a new message with the fields given as
arguments, but it automatically includes the current entity ID in the
FROM field, unless the FROM field is given in the arguments, in which
case it is not overriden.

=cut

sub newMsg {
  my $self=checkref(shift);
  my %fields=@_;
  if (!$fields{FROM}) {
    $fields{FROM}=$self->ID;
  }
  return AAFID::Message->new(%fields);
}

=head1 Responding to a CONNECT message

CONNECT messages are generated in two situations: whenever an entity
starts, a CONNECT CHILD message is sent to its "up" entity, to inform
it that it has started successfully. When an "up" entity starts a new
sub-entity, it sends a CONNECT PARENT message, so that the new entity
identifies who its "up" entity is.

In any case, the entity that receives a CONNECT message has to react
to it by creating a new record for the connecting entity in its 
C<OpenHandleIDs> parameter. 

=cut

sub message_CONNECT {
  my $self=checkref(shift);
  my $msg=checkref(shift, "AAFID::Message");
  my $entityid=$msg->{FROM};
  my $openhandles=$self->getParameter('OpenHandleIDs');
  my $entityrecord=$openhandles->{$entityid};
  my ($entityclass, $entitydesc)=split(/ /, $msg->{DATA}, 2);

  $self->Log("messages", "Got CONNECT message from $entityid\n");
  if ($entityrecord && $entityrecord->{_PendingConnect}) {
    $self->Log("messages", "I was expecting her. Checking consistency.\n");
    if (!defined($entityrecord->{Class}) || 
	($entityrecord->{Class} ne $entityclass)) {
      if ($entityrecord->{Class}) {
	$self->Log("errors", "WARNING: Class mismatch: I had '".
	    $entityrecord->{Class}.", she reports $entityclass. ".
	    "I'm keeping what I had.\n");
      }
      else {
	$self->Log("messages", "Updating Class name to '$entityclass'\n");
	$entityrecord->{Class} = $entityclass;
      }
    }
    if (!defined($entityrecord->{Description}) || 
	($entityrecord->{Description} ne $entitydesc)) {
      if ($entityrecord->{Description}) {
	$self->Log("messages", "She is giving me a new description ".
	    "'$entitydesc'. I had '".$entityrecord->{Description}."'. ".
	    "That's ok, it's her life.\n");
      }
      else {
	$self->Log("messages", "Setting Description to '$entitydesc'\n");
      }
      $entityrecord->{Description} = $entitydesc;
    }
    delete $entityrecord->{_PendingConnect};
    return undef;
  }
  elsif ($entityrecord) {
    $self->Log("errors", "I already have her, and was not expecting a CONNECT. Something is wrong. Doing nothing.\n");
    return undef;
  }
  else {
    $self->Log("messages", "She's new. Adding to the tables\n");
    $entityrecord=($openhandles->{$entityid}={});
    $entityrecord->{Class} = $entityclass;
    $entityrecord->{Description} = $entitydesc;
    # IMPORTANT: In this case, we assume that the same handle the message
    # came from is used to send messages. This may be wrong, but is
    # correct for TCP connections, which is the planned case where this
    # block will execute.
    $entityrecord->{Handle} = $msg->{_FROMHANDLE};
    return undef;
  }
}

=head1 Object identification

The C<ID> method simply returns the identifier of the entity. If it has
not been assigned, it creates and returns one, although it does not enter
it in the parameter table (that is to be done by the constructor).

=cut

sub ID {
  my $self=checkref(shift);
  my $id=$self->getParameter('EntityID');
  if ($id) {
    return $id;
  }
  else {
    my $plainhost;
    my $plainclass=ref($self);
    my $version=$self->getParameter('ClassVersion') || "0.00";
    my $number=$self->getParameter('InstanceNumber');
    $plainclass =~ s/^AAFID:://;
    ($plainhost=hostname) =~ s/\..*$//;
    return "$plainhost:$plainclass:$version:$number";
  }
}

=pod

The C<breakID> subroutine does the inverse process, breaking a provided
ID (or our own if no arguments are given) into its component parts, and
returning it as a hash reference. It does not necessarily have to be called
as an instance method, so it can be called by other programs.

=cut

sub breakID {
  my $self;
  my $id;
  # If first argument is a ref, assume we are being called as instance method.
  if (ref($_[0])) {
    $self=checkref(shift);
    $id=shift || $self->ID;
  }
  else {
    $id=shift || "";
  }
  if ($id =~ /^([^:]+):([^:]+(?:::[^:]+)*):([^:]+):(\d+)/) {
    return {Host 	=> $1,
	    Class 	=> $2,
	    Version	=> $3,
	    Number	=> $4
	   };
  }
  else {
    return undef;
  }
}

=head1 Running and stopping entities

All entities in the AAFID system are independent-running entities. Each
one of them is a separate process in the current implementation, but they
may be separate threads, or any other kind of running object.

Thus, after an entity object is created, we need a mechanism for making
it run, which should make the object start doing whatever it needs to do.
After this, we need a way of stopping (and possibly restarting) the 
entity at will. This will be necessary, for example, to allow a transceiver
to control several independent agents running under its control.

To this end, we have to establish some conventions. The ones we will use
for now are the following:

=over 4

=item *

Every entity must have an entry point routine called C<run()> which takes
no arguments (except the mandatory reference to itself that all instance
methods take in Perl) and starts the execution of the entity.

It is important that this subroutine sends a CONNECT message up before
starting to do anything. It is important that implementers of subclasses
of Entity remember to do this.

=cut

sub run {
    my $self=shift;
    my $class=ref($self);
    croak "AAFID::Entity: run() can only be called as an instance method"
	unless $class;
    $self->setupDefaultSignalHandlers;
    # Send a CONNECT message
    $self->sendConnectMessage;
    # Connect to the filters I requested, if any
    $self->connectFilters;
    # Template run() method, which does nothing.
    while(1) {
	$self->processInput(-1);
    }
}

=pod

Notice that the C<run> method is also responsible for calling the routine
that processes input to the entity, at regular intervals. See below for
the details on this.

=item *

Every entity must also be able to run as an independent program. For this,
we provide a hackish method called C<_EndOfEntity>, which should be the last
thing in the definition of an AAFID entity. This method checks whether the
file is being loaded from another program or being run as a standalone
program. If it is running standalone, it creates an instance of the current
class, and invokes the C<run> method. If not, it simply returns the value 1,
the customary value that has to be returned by Perl packages.

=cut

sub _EndOfEntity {
    my @c0=caller(0);
    my @c1=caller(1);

    if(@c1) {  # Being called from another program
	return 1;
    }
    else {     # Running standalone.
	my $obj=$c0[0]->new;
	$obj->run;
	return 1;
    }
}

=item *

Entities should respond to several signals as follows.

=over 4

=item USR1

Means "pause temporarily; do nothing if you are already paused".

=item USR2

Means "restart running; do nothing if you are already running".

=item TERM

Means "terminate gracefully".

=item HUP

Means "restart yourself".

=back 4

B<AAFID::Entity> provides very simple signal handlers for these purposes.
For B<USR1>, it simply sleeps the process until some other signal happens.
For B<HUP> it calls the C<run> method again, and for B<TERM> it simply
exits. These sample signal handlers are installed by the default C<run> 
method, so any descendant classes should be careful to set them appropriately.
The method for setting the default signal handlers is called
C<setupDefaultSignalHandlers>

=cut

sub DefaultHandler_USR1 {
    local $SIG{USR1}='IGNORE';
    sleep(0);
    $sleeping=1;
    while($sleeping) {
	Log("signals", "Sleeping...\n");
	sleep;
    }
}

sub DefaultHandler_USR2 {
    $sleeping=0;
    return;
}

sub DefaultHandler_TERM {
    exit(0);
}

sub DefaultHandler_HUP {
    $sleeping=0;
    Log("signals", 
	"Got HUP, but this signal handler is not yet implemented\n");
    # This does not work yet. run can only be called as an instance method.
    #run;
}

sub setupDefaultSignalHandlers {
  my $self=checkref(shift);
  $SIG{USR1} = \&DefaultHandler_USR1;
  $SIG{USR2} = \&DefaultHandler_USR2;
  $SIG{TERM} = \&DefaultHandler_TERM;
  $SIG{HUP}  = \&DefaultHandler_HUP;
}

=back 4

=head1 Terminating execution

At some point, the entity may need to finish its execution. If this is
done through the normal mechanisms, the C<Cleanup> routine will be
called on the object. This routine should do any necessary actions for
a graceful shutdown, and should be appropriately overriden by the
entity writer. It will be automatically passed a string (possibly empty)
that contains a description of the reason for the shutdown.

Internally, the C<_cleanup> subroutine is called, which calls the
user-provided C<Cleanup> subroutine, and then automatically generates
a DISCONNECT message and sends it both up and to any other entities
with which it is communicating.

The sending of the DISCONNECT message is actually done by a separate
subroutine C<sendDisconnectMessage>, so that it can be used by
subclass implementers. It takes an optional argument that gives a
description for the termination.

By default, the message is sent up with subtype of "CHILD". If a
second argument is given, it is interpreted as an entity ID, and the
DISCONNECT message is sent to that entity with a subtype of "PARENT".

=cut

sub sendDisconnectMessage {
  my $self=checkref(shift);
  my $desc=shift;
  my $to=shift;
  my $msg=$self->newMsg(TYPE		=> "DISCONNECT",
			SUBTYPE		=> ($to?"PARENT":"CHILD"),
			DATA		=> $desc);
  if ($to) {
    $self->Log("I/O", "Sending DISCONNECT message to $to: ".$msg->toString."\n");
    $self->sendMsgTo($msg, $to)
  }
  else {
    $self->Log("I/O", "Sending DISCONNECT message up: ".$msg->toString."\n");
    $self->sendReport($msg);
  }
}

sub Cleanup {
  my $self=checkref(shift);
  my $desc=shift;
  # Put your cleanup code here
}

sub _cleanup {
    my $self=checkref(shift);
    my $desc=shift;

    $self->Cleanup($desc);
    $self->sendDisconnectMessage($desc);
    exit(0);
}

=head1 Interpretation of messages and execution of commands

AAFID entities have the ability to receive messages and perform actions
based on them. Messages may be received either from "above" (this is,
from the controller entity) or, for controller entities (for example,
Transceivers and Monitors), from the sub-entities that are running at
any specific moment.

The general format of a message is as follows (as defined by the
B<AAFID::Message> class):

  <message type> <message subtype> <from> <to> <time> <data>

where the message types and subtypes that are recognized determine how
the data field is interpreted, and what actions are to be taken.

To implement the actions to be taken in response to different messages,
we define a set of methods called C<message_YYY>, where YYY is the 
message type to which that method will respond. When a message of type YYY
is received, that method will be invoked with one argument: an 
B<AAFID::Message> object containing the message just received. The
C<message_YYY> method must be defined with the message type part (the YYY)
in uppercase.

If the message method returns something, it must be an B<AAFID::Message>
object, which will be sent back to whoever sent the original message.

If the requested message method does not exist, the default method 
C<message_NoSuchMessageType> will be called, whose default action is to
produce an ERROR message.

One special case is the COMMAND message type, which indicates specific
actions to be taken by the receiving entity. A COMMAND message has
the following format:

  COMMAND <command name> <from> <to> <time> <parameters>

where B<<parameters>> is of the form I<key => value[,...]>, in the
traditional Perl hash syntax. The specific parameters that have to be
provided depend on the specific command being invoked.

Any Entity subclass can define its own set of commands, depending on
the functionality it needs. To define a new command, the class writer
simply has to define a method called C<command_XXX>, where XXX is the
name of the command to define. When the command is invoked, the subroutine
will be invoked with two arguments: a message object containing the
parsed message, and a hash (possibly empty) containing the parameters 
that were provided. The C<command_XXX> method must be defined with the
command name (the XXX part) in uppercase.

A command subroutine should normally not return anything (or, more
precisely, return C<undef>). If it returns something, it must be a
hash B<reference> containing parameter/value pairs, which will be sent
back to whoever sent the original command in a message of the
following format:

  COMMAND RESULT <from> <to> <time> Command=><command name>[, <values>]

where B<<command name>> will contain the command that triggered the 
response, and B<<values>> will be the hash returned by the command
method.

If a requested command is not found in the class, the method 
C<command_NoSuchCommand> will be invoked. The default implementation
provided by B<AAFID::Entity> simply ignores the command.

The B<AAFID::Entity> class provides a basic set of commands:

=over 4

=item STOP

Stop the entity by calling its C<Cleanup> method and exiting. It can
also be invoked as a message type. Automatically causes the sending
of a DISCONNECT message, both up and to any other entities that may
be registered in the C<OpenHandleIDs> parameter.

=item EVAL C<Code> => I<code>

The C<Code> parameters should contain valid Perl code which will be
executed in the context of the receiving entity. This is mostly for
experimentation purposes.

=item SET_PARAMS Param => I<value> ...

The parameters specify a set of parameter/value pairs that will be set
in the entity using the C<setParameters> method.

=item GET_PARAMS Params => "name[, name ...]"

Returns a COMMAND RESULT message containing the requested parameters
and their values in C<Name =E<gt> Value> form.

=item DUMP_YOURSELF

Return a string representation of the object, as produced by the standard
B<Data::Dumper> package. Internally, an entity is represented as a hash
that contains the parameters of the entity.

=back 4

The C<processInput> subroutine is the general one that waits for input
on a set of input handles, and processes any messages that come on any
of them by reacting correspondingly.

If no parameters are passed to it, C<processInput> does no blocking,
and only reads and processes the next available message, if any. If a
numeric parameter is passed, it will be interpreted as a timeout, in
seconds, to wait for a message to become available before
returning. Finally, if C<-1> is passed as a timeout, the call will
block.

C<processInput> automatically adds C<STDIN> to the Entity
C<OpenHandles> and C<OpenHandleIDs> parameters (the adding to
C<OpenHandles> is actually done by C<AAFID::Comm::nextMsg>), so those
parameters must be initialized empty.

If the entity defines the subroutines C<PreMessageHook> and 
C<PostMessageHook>, they will be called for all messages destined
for us before the message is processed, and afterwards. C<PreMessageHook>
will be called with the message as an argument, and C<PostMessageHook>
will be called with the message and the result from the subroutine
that processed it as arguments.

=cut

sub processInput {
  my $self=checkref(shift);
  # Get the appropriate timeout setting.
  my $timeout=shift;
  if (defined($timeout)) {
    local $^W=0;
    $timeout+=0;
    if ($timeout==-1) {
      $timeout=undef;
    }
  }
  else {
    $timeout=0;
  }
  # Get the message.
  my $msg=AAFID::Comm::nextMsg(Timeout => $timeout, 
			       Select => $self->getParameter('OpenHandles'));
  # Nothing, return
  return undef if !defined($msg);
  # If got a close on STDIN, exit gracefuly.
  if($msg==-1) {
    $self->Log("I/O", "Got close on STDIN, exiting\n");
    $self->_cleanup("Close on STDIN");
    return undef;
  }
  if($msg && $msg->{_FROMSTDIN} && $msg->{FROM}) {
    my $oh=$self->getParameter('OpenHandleIDs');
    if(!defined($oh->{$msg->{FROM}})) {
      $self->Log("I/O", "Creating record for ".$msg->{FROM} .
		 " in OpenHandleIDs\n");
      $oh->{$msg->{FROM}} = {};
      $oh->{$msg->{FROM}}->{_PendingConnect} = 1;
      $oh->{$msg->{FROM}}->{_isUp} = 1;
    }
    if (!defined($oh->{$msg->{FROM}}->{Handle})) {
      $self->Log("I/O", "Associating ".$msg->{FROM}." to STDOUT\n");
      $oh->{$msg->{FROM}}->{Handle}=IO::File->new_from_fd(\*STDOUT, "w");
      $oh->{$msg->{FROM}}->{Handle}->autoflush(1);
      $oh->{$msg->{FROM}}->{_isUp} = 1;
    }
  }
  if($msg) {
    my $msgtype=uc($msg->{TYPE});
    my $result;
    if ($self->isForMe($msg)) {
      $self->Log("debug", 
		 "Got a $msgtype message, trying to call message_$msgtype\n");
      if ($self->can("PreMessageHook")) {
	$self->PreMessageHook($msg);
      }
      my $msgsub=$self->can("message_$msgtype");
      if(!$msgsub) {
	$result=$self->message_NoSuchMessageType($msg);
      }
      else {
	$result=&{$msgsub}($self, $msg);
      }
      if ($self->can("PostMessageHook")) {
	$self->PostMessageHook($msg, $result);
      }
      if ($result) {
	if(ref($result) ne "AAFID::Message") {
	  croak "Message_${msgtype} generated a non-Message result!";
	}
	$self->sendMsgTo($result, $msg->{FROM});
      }
    }
    else {
      $self->relayMessage($msg);
    }
  }
  return $msg;
}

sub relayMessage {
  my $self=checkref(shift);
  my $msg=checkref(shift, "AAFID::Message");
  my $msgto=$msg->{TO};
  my $msgtype=uc($msg->{TYPE});
  $self->Log("debug",
	     "Got a $msgtype message, but not for me, it's for $msgto\n");
  $self->sendMsgTo($msg, $msgto);
}

# Determine if the message is for me. A TO field of "-" acts as a wildcard.

sub isForMe {
  my $self=checkref(shift);
  my $msg=checkref(shift, "AAFID::Message");
  my $msgto=$msg->{TO};
  my $oh=$self->getParameter('OpenHandleIDs');
  if ( ($msgto ne "-") && 
       (lc($msgto) ne lc($self->ID)) && 
       !exists $oh->{$msgto}) {
    $self->Log("debug", "The message is for $msgto, but I don't know her, " .
                 "so I'll take it.\n");
  }
  return ( ( lc($msgto) eq lc($self->ID) ) or
	   ( $msgto eq "-" ) or 
	   ( !exists $self->getParameter('OpenHandleIDs')->{$msgto}) );
}

# Definition of subroutines for message types.

# Default method if unknown message.
sub message_NoSuchMessageType {
    my $self=checkref(shift);
    my $msg=checkref(shift, "AAFID::Message");
    $self->Log("messages", "Undefined message type: ".$msg->toString."\n");
    # Do nothing.
    return undef;
}

sub message_COMMAND {
    my $self=shift;
    ref($self) or croak "message_COMMAND can only be called as an instance method";

    my $msg=shift;
    # Get the command name
    my $cmdname=uc($msg->{SUBTYPE});
    $self->Log("messages", "Got a $cmdname command\n");
    # Get the parameters
    my %paramhash=();
    my $paramstring=$msg->{DATA};

    if($paramstring) {
	$self->Log("debug", "Evaluating %paramhash=($ {paramstring})\n");
	eval "%paramhash=($ {paramstring})";
	if($@) {
	    chomp $@;
	    $self->Log("messages", "Error evaluating command parameters: $@\n");
	    my $nmsg=$self->newMsg(TYPE	=> "COMMAND",
				   SUBTYPE=> "ERROR",
				   DATA	=> "ParamError => $@");
	    $self->sendMsgTo($nmsg, $msg->{FROM});
	    return;
	}
    }
    # Execute the command
    my $result;
    $self->Log("messages", "Trying to call command_$cmdname\n");
    my $cmdsub=$self->can("command_$cmdname");
    if(!$cmdsub) {
	chomp $@;
	$result=$self->command_NoSuchCommand($msg, %paramhash);
    }
    else {
	$result=&{$cmdsub}($self, $msg, %paramhash);
    }
    if ($result) {
	$self->Log("debug", "command_${cmdname} produced a result: ".Dumper($result)."\n");
	# TODO: How to detect if %result is indeed a hash?
	$$result{Command}=$cmdname;
	my $datadump=Dumper($result);
	$datadump =~ s/^\s*\{(.*)\}\s*/$1/;
	my $nmsg=$self->newMsg(TYPE	=> "COMMAND",
			       SUBTYPE	=> "RESULT",
			       DATA	=> $datadump);
	# TODO: make this send adequately to whoever sent the command
	$self->Log("debug", "Sending RESULT message to $msg->{FROM}: ".$msg->toString."\n");
	$self->sendMsgTo($nmsg, $msg->{FROM});
    }
    return undef;
}

sub command_NoSuchCommand {
    my $self=shift;
    my $msg=shift;
    $self->Log("messages", "Command " . $msg->{SUBTYPE} . " not defined: "
	. $msg->toString."\n");
    # Do nothing here
    return undef;
}

sub message_STOP {
  my $self=checkref(shift);
  $self->command_STOP(@_);
}

# Definition of soubroutines for different commands.

sub command_STOP {
    my $self=shift;
    ref($self) or croak "command_STOP can only be called as an instance method";
    # Don't even look at the arguments.
    $self->_cleanup("STOP command or message");
    exit(0);
}

sub command_EVAL {
    my $self=checkref(shift);
    # For now, blindly eval this.
    my ($msg, %params)=@_;
    $self->Log("debug", "Eval'ing ".$params{Code}."\n");
    eval $params{Code};
    if($@) {
	chomp $@;
	$self->Log("debug", "Error in eval: $@\n");
	return {ErrorMessage=>$@};
    }
    return undef;
}

sub command_SET_PARAMS {
    my $self=checkref(shift);
    my ($msg, %params)=@_;
    $self->setParameters(%params);
    return undef;
}

sub command_GET_PARAMS {
  my $self=checkref(shift);
  my ($msg, %params)=@_;
  if ($params{Params}) {
    my @params=split(/[,\s]+/, $params{Params});
    my %paramvalues=$self->getParameters(@params);
    return \%paramvalues;
  }
}

# This may be useful for debugging

sub command_DUMP_YOURSELF {
  my $self=checkref(shift);
  my $result=Dumper($self);
  return {Me => $result};
}

=head1 Sending message "up"

The Entity class includes this function simply to make it easier for
Entity writers to send messages "up" (this is, to the parent entity).
This is simply a gateway to the corresponding method in
B<AAFID::Comm>.

=cut

sub sendReport {
  my $self=shift;
  ref($self) or croak "AAFID::Entity::sendReport can only be called as an instance method";
  my $msg;
  foreach $msg (@_) {
    if(ref($msg) ne "AAFID::Message") {
      $self->Log("errors", "sendReport: got a non-Message argument: $msg");
      next;
    }
    else {
      $self->Log("debug", "Sending message up: ".$msg->toString."\n");
      AAFID::Comm::sendMsgUp($msg);
    }
  }
}

=head1 Sending a message to an arbitrary entity.

This subroutine takes as arguments a message and an entity identifier. It
looks up the handle associated with that identifier, and send the message
to it.

=cut 

sub sendMsgTo {
  my $self=shift;
  ref($self) or croak "sendMsgTo can only be called as an instance method";
  my $msg=shift;
  my $whoID=shift;
  my $oh=$self->getParameter('OpenHandleIDs');
  my $h;
  if ($oh->{$whoID} && ($h=$oh->{$whoID}->{Handle})) {
    $self->Log("debug", "Sending message to $whoID: ".$msg->toString."\n");
    AAFID::Comm::sendMsgTo($msg, $h);
  }
  else {
    $self->Log("errors", "Invalid ID: $whoID\n");
  }
}

=head1 Filter support

If an entity requests filters through its C<FiltersNeeded> parameter,
the controller entity that starts it will take care of starting the
filters before running the entity, and will put the information necessary
to contact each filter in the C<FilterPaths> parameter, with each key
representing the filter name and each element containing the path of
the Unix domain socket through which the filter can be contacted.

Once the entity starts, it has to contact the filters, and needs a way
of reading data from the filters and of telling the filter the pattern
it wants to match for the data it receives.

The C<connectFilters> subroutine opens sockets to all the filters
given in C<FilterPaths>, and sends to each one of them the initial
pattern as given in C<FiltersNeeded>.

=cut

sub connectFilters {
  my $self=checkref(shift);
  my $fp=$self->getParameter('FilterPaths', {});
  my $f=$self->getParameter('FilterSockets', {});
  my $fname;
  $self->Log("I/O", "Contacting filters\n");
  foreach $fname (keys %$fp) {
    my $fpath=$fp->{$fname};
    my $sock=$self->newFilterClientSocket($fpath);
    if ($sock) {
      $self->Log("I/O", "Opened socket $fpath to filter $fname.\n");
      $f->{$fname}=$sock;
      $self->sendConnectMessage($sock);
      $self->setFilterPattern($fname, 
			      $self->getParameter('FiltersNeeded')->{$fname});
    }
    else {
      $self->Log("errors", 
		 "Error opening socket $fpath to filter $fname: $!\n");
    }
  }
}

sub newFilterClientSocket {
  my $self=checkref(shift);
  my $addr_sock=shift;
  if (!$addr_sock) {
    $!="No socket path given";
    return undef;
  }
  my $sock=IO::Socket::UNIX->new(Type           => SOCK_STREAM,
                                 Peer           => $addr_sock,
				);
  $sock->autoflush(1) if $sock;
  return $sock;
}

=pod

The C<setFilterPattern> subroutine takes as argument a filter name and a
hash containing the pattern that needs to be set for that filter, and 
sends the appropriate message.

=cut

sub setFilterPattern {
  my $self=checkref(shift);
  my $fname=shift;
  my $patref=shift || {};
  my $pat=hashToStr($patref);
  if (!$fname) {
    $self->Log("errors", "setFilterPattern: no filter name given.\n");
    return;
  }
  else {
    my $socks=$self->getParameter('FilterSockets', {});
    if ($socks->{$fname}) {
      $self->Log("debug", "Sending pattern ($pat) to filter $fname.\n");
      my $msg=$self->newMsg(TYPE    => 'COMMAND',
			    SUBTYPE => 'SETPATTERN',
			    DATA    => $pat
			   );
      AAFID::Comm::sendMsgTo($msg, $socks->{$fname});
    }
  }
}

=pod

To read from a filter, the subroutine C<readFilter> is used. It gets
as argument the name of the filter, and returns the next line of input
from the filter, if one is available. If there are no lines available
or an error occurs, it returns C<undef>. If an error occurs, it also
sets the C<$!> variable.

=cut

sub readFilter {
  my $self=checkref(shift);
  my $fname=shift;
  if ($fname) {
    my $socks=$self->getParameter('FilterSockets', {});
    if ($socks->{$fname}) {
      my $line=AAFID::Comm::nextLine(Handle => $socks->{$fname},
				     Timeout => 0,
				    );
      return $line;
    }
    else {
      $self->Log("errors", "Socket for filter $fname is not open.\n");
      $!="Socket for filter $fname is not open";
      return undef;
    }
  }
  else {
    $self->Log("errors", "No filter name given for read.\n");
    $!="No filter name given for read";
    return undef;
  }
}

=head1 Parameter manipulation

Identification and some behavior control for entities is provided
through named parameters. The names of the parameters are 
case-insensitive (internally, they are all converted to lowercase).

=head2 Accessing parameters through the standard methods

In the current implementation, these parameters are stored in the
object itself, which is represented as a hash. However, it is not
recommended that the parameters are accessed directly through the
hash. Instead, the C<setParameters> and C<getParameters> methods must
be used. C<setParameters> receives a hash with S<key => value>
elements, and returns the object reference itself. C<getParameters>
receives a list of parameter names, and returns a hash with the
requested parameters as keys, and their values (or C<undef> if the
requested parameter does not exist).

Another method used to access the parameters is C<getParametersList>,
which receives the same type of arguments as C<getParameters>, but 
instead of returning a hash, returns a list with only the values of
the requested parameters, in order, and with C<undef> in the place of
non-existant parameters.

For reducing typing errors, C<setParameter> behaves exactly like
C<setParameters>, and C<getParameter> (in singular) does the right
thing: returns not a hash, not an array, but the value of the
parameter whose name is passed as its first argument. Additionally,
if a second parameter is passed to C<getParameter>, it is used
to initialize the parameter if it is not defined.

Finally, the C<getParametersRef>, C<getParameterRef> and 
C<getParametersListRef> subroutines behave exactly like their non-Ref
counterparts, but instead of returning the values of the requested
parameters, return references ot them, so that the caller can store them
and later modify the parameters directly (not recommended), or read the
values dynamically in case they change through some other mechanism
(useful).

=cut

sub setParameters {
    my $self=shift;
    my %params=@_;
    my $k;
    foreach $k (keys %params) {
      $self->{lc($k)}=$params{$k};
    }
    return $self;
}

sub setParameter {
    return setParameters(@_);
}

sub getParameters {
    my $self=shift;
    my $k;
    ref($self) 
	or croak "getParameters can only be called as an instance method";
    my %returnedparams=map { $_ => $self->{lc($_)} } @_;
    return %returnedparams;
}

sub getParameter {
  my $self=shift;
  my $param=shift;
  my $value=shift;
  # Get the value of the parameter
  my $paramvalue=$self->{lc($param)};
  # If it is undefined and a default value is used, initialize it.
  if ( (!defined($paramvalue)) && defined($value) ) {
    $self->setParameter($param => $value);
    $paramvalue=$self->{lc($param)};
  }
  return $paramvalue;
}

sub getParametersList {
    my $self=shift;
    ref($self)
	or croak "getParametersList can only be called as an instance method";
    return map { $self->{lc($_)} } @_;
}

sub getParametersRef {
    my $self=shift;
    ref($self) 
	or croak "getParameters can only be called as an instance method";
    return map { $_ => \$self->{lc($_)} } @_;
}

sub getParameterRef {
    return (getParametersListRef(@_))[0];
}

sub getParametersListRef {
    my $self=shift;
    ref($self)
	or croak "getParametersList can only be called as an instance method";
    return map { \$self->{lc($_)} } @_;
}

=head2 Accessing parameters through a hash

An easier but less powerful method for accessing parameters is also
available. When an B<AAFID::Entity> or a subclass of it is instantiated,
a hash called C<Params> is automatically created, which can be used
from anywhere in the object to access the parameters. Readings and
assignments to elements of the C<Params> hash are equivalent to calls
to C<getParameter> and C<setParameter>, respectively.

It is important to note some points:

=over 4

=item *

The C<%Params> array is a B<class> variable, not an instance one. This means
that it is shared by all the objects of the same class that inhabit a single
Perl process. This is not a problem in the current implementation because
each entity is a separate process.

=item *

If references to the parameters are needed, the C<getParametersRef>,
C<getParameterRef> or C<getParametersListRef> methods need to be used,
because if references are accessed through the C<%Params> hash, references
to copies of the actual parameters will be obtained.

=back 4

To implement the possibility of accessing parameters through a hash,
the methods C<TIEHASH>, C<FETCH> and C<STORE> are defined, and the 
tying operation is performed in the C<setTie> method, which is called
during initialization.

=cut

sub TIEHASH {
  my $class=shift;
  my $object=checkref(shift);
  bless $object, $class;
  return $object;
}

sub FETCH {
  my $self=checkref(shift);
  my $key=shift;
  my $value=$self->getParameter($key);
  $self->Log('params', "FETCH called with key '$key', returning '" .
	     ($value?"$value":"undef") . "'.\n");
  return $value;
}

sub STORE {
  my $self=checkref(shift);
  my $key=shift;
  my $value=shift;
  $self->Log('params', "STORE called with key '$key', value '$value'.\n");
  $self->setParameter($key => $value);
  return $value;
}

sub setTie {
  my $self=checkref(shift);
  my $class=ref($self);
  my $varname="${class}::Params";
  no strict 'refs';
  tie %$varname, $class, $self;
  return $self;
}

=head1 Standard entity parameters

The parameters that are understood by all entities (the B<AAFID::Entity>
class and all of its subclasses) are:

=over 4

=item ClassCount

A reference to the number of entities of this class that currently
exist.  To access it, you would have to do something like
C<${getParameter("ClassCount")}>.

=item InstanceNumber

The number assigned to the current instance.

=item ClassID

The class name of the entity.

=item Description

Optional parameter which gives a textual description of the Entity,
for identification purposes.

=item ClassVersion

A version number of the form B<major.minor>, assigned by the class writer
and declared in the global variable C<$VERSION>. If using RCS, a good way
of making the version number coincide with the RCS revision number is 
inserting this:

	# The following keeps up with the RCS version number. The
	# self-assignment keeps the -w switch from complaining
	# (because $VERSION may not be used here, but it is used in
	# our base class).
	$VERSION = do { my @r = (q$Revision: 1.36 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; $VERSION = $VERSION;

=item EntityID

A unique identifier for the entity, of the form 
C<hostname:ClassID:Version:InstanceNumber>.

=item OpenHandles

An B<IO::Select> object that contains the handles of all the currently
open input handles. It is initialized to an empty object, but eventually
at least C<STDIN> will be added to it. This is done automatically by
the C<AAFID::Comm::nextMsg> subroutine, to which this object must be
passed in every call.

It will also contain the reading end of the pipes of any sub-entity
that this entity is controlling.

This parameter is handled automatically by the Entity class, the entity
writer does not have to worry about it.

=item OpenHandleIDs

A reference to a hash that relates entity IDs to information about
them. This information is stored under different keys: 

=over 4 

=item C<Handle> 

Contains a reference to the file handle object that is used to send
information to the entity.

=item C<Description> 

Contains the entity's descriptive text, if any. %'

=item C<Class>

Contains the entity's class name. %'

=item C<_PendingConnect>

If true, indicates that the entity corresponding to this record was
started by us, but has not yet sent a CONNECT message to announce that
it is alive. It also happens when a message comes from standard input
from an entity that had not yet been registered. It is automatically
registered, and this flag is set in the hopes that the message itself
will be a CONNECT PARENT message.

=item C<_isUp>

If true, indicates that the entity corresponding to this record
is our "up" entity.

=back 4

This parameter, as well as C<OpenHandles>, are handled completely by
the Entity base class. The entity writer does not have to worry about
initializing them.

=item FiltersNeeded

(hash reference) If the entity needs any filters to be active for its
operation, this parameter contains that information. Each key in the
hash is the name of a filter, and the element contains a hash reference
that contains the initial pattern that will be sent to that filter as
filtering criteria for data that the entity needs.

For example, an agent that needs to get data from the
B<NetworkAccesses> filter, and that only needs to see the data from
C<telnet> requests, could specify the following in its C<%PARAMETERS>
hash:

  FiltersNeeded => {
		    NetworkAccesses => {
					Service => 'telnetd'
				       }
		   }
=item FilterPaths

(hash reference) Contains information that allows the Entity to
contact each of the filters it needs. Each key is a filter name, and
the corresponding element contains the path in which the Unix-domain
socket for the filter is located, and through which the entity can
contact it. This parameter is set by the controller entity that starts
the entity.

=back 4

=cut

=head1 Standard entity messages and commands

TODO

=cut

_EndOfEntity;

#
# $Log: Entity.pm,v $
# Revision 1.36  1998/09/28 15:39:49  zamboni
# Made some changes in the documentation that were suggested by
# Stephanie Miller on her evaluation of the prototype. Also, applied a
# patch provided by Frederic Dumont that adds some error checking to the
# agents and the initialization of an Entity (it checks for Init
# returning 'undef') and corrects a color in the GUI ("darkyellow"
# apparently does not exist on 8-bit displays). Finally, made the
# default value of LogFile in config/AAFID be /tmp/aafid.log.
# Added Linda Jessen to the CREDITS file.
#
# Revision 1.35  1998/09/09 09:04:53  zamboni
# * Entity.pm: made readLine return undef on both EOF and
#   error, but to also set $! on error.
#
# Revision 1.34  1998/09/07 17:35:20  zamboni
# Added support for filters!
#
# * Monitor.pm: Replaced all uses of AAFID::Message->new by uses
#   of $self->newMsg.
#
# * Filter.pm: Cleaned up the code and the comments, and added some
#   extra features, such as default versions of makefield and
#   makeline that make it easier to define new filters if they read
#   data where the fields are space-separated. Also, added some
#   error checks here and there.
#
# * Entity.pm: Added filter support
#   - Added descriptions for FiltersNeeded and FilterPaths parameters.
#   - Modified getParameter so that if a second argument is given, it is
#     used to initialize the parameter in case it is undefined.
#   - Added subroutines connectFilters, newFilterClientSocket,
#     setFilterPattern
#   - Added subroutine newMsg.
#
# * ControllerEntity.pm: Added filter support.
#   - Some general cleanup (removing stray commented code, etc.)
#   - Rewrote some sections for clarity (in invoke, _invoke,
#     _instantiate_and_run)
#   - Modularized loadmodule. Created new subroutines:
#       _loadmodule_error
#       _loadmodule_success
#   - Modularized _instantiate_and_run, creating new subroutine
#       _fork_and_run.
#   - Added the setupFilters subroutine, which takes care of loading
#     all the filters required by an entity. It is complemented by
#     an augmented message_CONNECT, which detects when a filter is
#     connecting and does the necessary updating.
#
# * Comm.pm: Removed an annoying debug message from bufferInput.
#
# * Agent.pm: Added code to contact the filters.
#
# Revision 1.33  1998/08/26 22:36:56  zamboni
# Reimplemented the reading mechanisms to use non-buffered I/O. This fixes
# problems that ocurred when mixing buffered I/O with the use of the Select
# object. Now the AAFID::Comm library keeps an internal buffer of lines
# read, and allows to retrieve them through nextMsg (interpreting them
# as AAFID messages, as usual) and through a new subroutine called nextLine,
# which returns the line read.
#
# Also, all the reading routines (nextMsg and now nextLine) return only
# the next available element. I had changed nextMsg so that it returned an
# array of messages, but that was a bad decision, I think, so we went back
# to the old behavior.
#
# Moved the RCS log to the end of each file.
#
# Now nextMsg and nextLine take named parameters for Timeout, Handle and
# Select object.
#
# Entity.pm and Monitor.pm were modified to go back to the behavior of Comm::nextMsg
# returning a single message. They had been modified to process a list of messages,
# but this was seen to be a bad decision.
#
# Also, calls to nextMsg were modified to make use of the new syntax, which requires
# named parameters.
#
# Revision 1.32  1998/08/11 16:24:00  zamboni
# - Made processInput use the new feature of AAFID::Comm::nextMsg that returns
#  a list of messages instead of a single message. Now processInput processes
#  at once all the pending messages, instead of only the first one.
#
# Revision 1.31  1998/06/29 20:11:23  zamboni
# Added copyright message
#
# Revision 1.30  1998/06/26 21:28:09  zamboni
# - Modified the constructor (sub new) to recursively incorporate all the
#   %PARAMETERS hashes of all the subclasses and the current class into
#   the object's parameters. This makes it unnecessary now to do any
#   manual additional incorporation, as was done previously in the _init
#   subroutine of Agent.pm. This also makes the interaction with the
#   configuration mechanism (AAFID::Config) cleaner, because now the
#   AAFID::Config::configure subroutine can be safely called before
#   _init, knowing that no default parameters will be set in _init
#   any more.
#
# - Changed the order to first set all the standard parameters (from the
#   %PARAMETERS hashes) and then to set the entity-specific parameters
#   such as entity counter, entity ID, etc. (previously it was the
#   other way around).
#
# - Added the use of AAFID::Config to get the log file name, as well as
#   the log categories that have to be activated. Now the constructor
#   of Entity registers and activates all the log categories.
#
# Revision 1.29  1998/06/25 20:48:52  zamboni
# - Made the entity parameters be case-insensitive. Also, modified
#   getParameters to not return the result of the map directly (which
#   could return a hash with repeated elements if the same parameter
#   is requested twice), but to assign that to a temporary internal hash
#   (which eliminates any repeated elements) and return the temporary
#   hash.
#
# - Made Entity a subclass of AAFID::Config and added a call to
#   $self->configure in the constructor.
#
# Revision 1.28  1998/06/17 19:47:38  zamboni
# - Added TIEHASH, STORE, FETCH and setTie methods to create a %Params hash
#   during object creation that allows access to the parameters through
#   that hash. A call to setTie was inserted in the constructor.
#
# Revision 1.27  1998/05/07 06:04:06  zamboni
# - Added relayMessage, which is called when isForMe returns false to
#   send the message to someone else. The purpose of making it into a
#   separate method is that it can then be overriden by subclasses (for
#   example, Monitor) to do the appropriate thing.
#
# Revision 1.26  1998/05/07 05:18:35  zamboni
# - Added the use of the two methods PreMessageHook and PostMessageHook.
#   If an entity that is a subclass of AAFID::Entity defines any (or
#   both) of these methods, they will be called before and after a
#   received message is processed, respectively, but only if the message
#   is destined for the entity (as reported by isForMe).
#
# - Made it possible to call breakID as a class method
#   (AAFID::Entity::breakID), and not only as an instance method.
#
# Revision 1.25  1998/05/03 03:36:25  zamboni
# - Corrected a condition check in isForMe that made it print an "I don't
#   know her" message when the message is for us.
#
# Revision 1.24  1998/05/02 19:51:38  zamboni
# POD correction.
#
# Revision 1.23  1998/04/27 15:09:29  zamboni
# - Added a command GET_PARAMS that receives an argument called "Params" that
#   itself contains a comma-and-space-separated list of parameter names
#   that are to be returned.
# - Added documentation for the GET_PARAMS and DUMP_YOURSELF commands.
# - Made only "messages" and "errors" log categories active by default.
#
# Revision 1.22  1998/03/17 06:39:37  zamboni
# - Changed cleanup to Cleanup to be consistent with Init.
#
# Revision 1.21  1998/03/13 16:59:06  zamboni
# - Added $VERSION and %PARAMETERS.
# - Made $VERSION and %PARAMETERS set correctly when the constructor is
#   called from a subclass. Previously, only the values defined in AAFID::Entity
#   were used. Now, those are used as default, but are then overriden by any
#   subclass-provided parameters.
# - Added description of how to set $VERSION to the documentation part.
#
# Revision 1.20  1998/03/13 01:47:53  zamboni
# Made $VERSION keep up with the RCS revision number.
#
# Revision 1.19  1998/03/06 07:10:55  zamboni
# - Made processInput return the message that was received.
# - Made $self->ID check if the ID has not been assigned, and if so, build
#   and return it. Then made the constructor use $self->ID instead of
#   building the ID there.
# - Added breakID subroutine.
# - Added "use strict"
# - Corrected problems pointed out by strict module. One not-so-nice "problem"
#   detected was the use of barewords for getParameter, as in
#   $self->getParameter(Foo), which had to be changed to
#   $self->getParameter('Foo') on strict's complain. Less convenient, but
#   safer, I guess.
# - Made processInput recognize when the message is for the current entity, and
#   if not, relay it appropriately. "For the current entity" means that the
#   TO field of the message contains either the ID of the current entity or
#   the wildcard "-".
# - Made it even more intelligent. If the message is not for the current
#   entity, but the entity to which it is addressed does not exist, we
#   take it.
#
# Revision 1.18  1998/03/05 03:32:41  zamboni
# - Added the command DUMP_YOURSELF
#
# Revision 1.17  1998/03/04 22:07:42  zamboni
# - Moved the exit(0) to inside _cleanup. This way, subclasses (such as
#   ControllerEntity) can override it to avoid exiting immediately when
#   a cleanup is called.
# - Removed the exit(0) from processInput after the call to _cleanup, and
#   replaced it with a return undef. See above.
# - Modified message_STOP to call $self->command_STOP and not simply
#   command_STOP, because otherwise it always calls Entity::command_STOP,
#   even if it has been overriden in the class it is being called from.
#
# Revision 1.16  1998/03/04 19:43:44  zamboni
# Removed spurious log message from sendMsgTo (it printed the Dump of the
# handle to which it is sending the message, which was unnecessary and ugly).
#
# Revision 1.15  1998/03/04 06:31:48  zamboni
# Did general debugging and cleaning up. Seems to mostly work now.
# Did the following changes in the entity IDs:
#   - 'AAFID::' is removed from the class name, if it contains it.
#   - The host name is reduced to only the host part, domain name is removed.
# TODO: These are just for the sake of space and readability of the log
# files. These changes (particularly the removing of the domain name)
# should be undone later.
#
# Revision 1.14  1998/02/26 21:22:33  zamboni
# Made sendDisconnectMessage() take two optional parameters: first, the reason
# for the disconnect, which will be sent literally in the DATA field of
# the message, and second, an entity ID to which the message must be sent. If
# this ID is not provided, the message is sent up with a subtype of CHILD. If
# it is there, it is sent to that entity with a subtype of PARENT.
#
# Revision 1.13  1998/02/26 21:00:48  zamboni
# - Added message_CONNECT. This was going to originally be in ControllerEntity,
#   but then I realized that ControllerEntities should also send a CONNECT
#   message down to their children, and thus any entity needs to be able
#   to react to it.
# - Changed all uses of OpenHandleIDs to its new structure, in which the
#   parameter is a reference to a hash keyed by entity ID, and each element
#   in turn is a hash with keys for different pieces of information about
#   the entity, such as Class, Description or Handle.
#
# Revision 1.12  1998/02/26 05:31:25  zamboni
# Changed the references to $msg->{FROMSTDIN} to $msg->{_FROMSTDIN}.
#
# Revision 1.11  1998/02/26 04:57:10  zamboni
# Disabled debugging output by default.
#
# Revision 1.10  1998/02/25 20:50:29  zamboni
# Changed a bit of wording in documentation.
#
# Revision 1.9  1998/02/25 19:28:38  zamboni
# Added a comment about sendConnectMessage in run().
#
# Revision 1.8  1998/02/25 19:26:57  zamboni
# Moved the sending of the CONNECT message from the constructor to the
# run() method. This is because when a new entity is instantiated from
# a controller entity, it will not yet be running on a separate process,
# and thus the message would go to the standard output of the controller
# entity. When the run() method is invoked, everything will be set up
# correctly.
#
# Revision 1.7  1998/02/25 07:20:20  zamboni
# - Added functions to get references to the parameters (getParametersRef,
#   getParameterRef, getParametersListRef)
# - Added section on standard entity parameters at the end.
# - Added internal method _init which is called from the constructor instead
#   of calling Init directly. _init in this class simply calls Init, but
#   in other classes it may be used to do some standard initialization
#   before (or after) calling the user-provided Init.
# - Made processInput take an optional argument to specify a timeout.
# - Corrected a number of minor errors.
#
# Revision 1.6  1998/02/23 05:41:06  zamboni
# First apparently fully working version. Correctly processes messages
# and commands.
#
# Revision 1.5  1998/02/18 20:09:26  zamboni
# First version that should kind of work by being able to receive messages
# through stdin, and generate output through stdout. Of course, it needs
# supporting functions from AAFID::Comm to do that, but I'll work on it now.
#
# Revision 1.4  1998/02/18 05:20:06  zamboni
# First checkin of the new version of Entity, using stdin/stdout for
# communication, and many other changes.
#
# Revision 1.3  1998/01/28 21:58:12  zamboni
# Changed the message printed by run().
#
# Revision 1.2  1998/01/27 03:33:02  zamboni
# Seems to work, mostly. The USR1 signal handler is looping once before
# stabilizing in the sleep (like if it was receiving two signals, or something),
# but it pauses just fine. My biggest problem now is the HUP signal handler.
# The problem is that I need to call the run method again, but it can only
# be called as an instance method. Thus, it cannot be called from the signal
# handler, because at that point we do not have information about the object
# we are in. I have to give that a little more thought.
#
# Revision 1.1  1998/01/26 23:30:52  zamboni
# Initial revision
#
