#!/usr/local/perl/perl -w
#
# AAFID project, COAST Laboratory, 1998.
#
# Susana Soriano, August 11, 1998.
#
# $Id: Filter.pm,v 1.3 1998/09/09 09:05:37 zamboni Exp $
#

package AAFID::Filter;

use AAFID::Entity;
use AAFID::Common;
use Data::Dumper;
use Socket;
use Carp;
use strict;
use IO::File;
use vars qw(
	    $waitedpid
	    $self
	    @ISA
	    $NAME
	    $Server
	    $Client
	    $FILEF	
	    $PERIOD
	    $VERSION
	    %PARAMETERS
	   );

@ISA=qw(AAFID::Entity);

# The following keeps up with the RCS version number. The self-assignment
# keeps the -w switch from complaining (because $VERSION may not be used
# here, but it is used in our base class).
$VERSION = do { my @r = (q$Revision: 1.3 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; $VERSION = $VERSION;

#Default parameters.
%PARAMETERS=(
	     CheckPeriod 	=> 5,
	     Description		=> "Base Filter class",
	    );

=pod

This is the basic structure of a filter. In this program, there are
defined the basic subroutines. The specifications depending on the
logfile must be defined in an independent package which has to be a
subclass of this filter. The structure of a filter contains the
subroutines that open a logfile if it is required, read the pattern(s)
that the agent gives to the filter, establish a UNIX-domain socket
which connects the agent with the filter and send the information
wanted to the agent.

=head1 Joining an Agent, a Filter and a Logfile.

Sometimes the information that is required for the Agent is stored in
a logfile. In fact, this is the general case and it is implemented as
a default case. The subroutine C<Init_log> takes care of opening the log
file and storing it in the C<MyFileHandle> parameter. The file opened is
given by the C<LogFile_Name> parameter, which must be defined by each
filter.

In the case that the information is not stored in a logfile (or at
least not in a single log file), the filter must redefine the
C<Init_log> subroutine accordingly. The filter opens a logfile if it is the
default case.

=cut

sub Init_log {
  my $self=checkref(shift);
  #Filter opens the file logfile if it is defined.
  if (defined ($FILEF)) {
    $self->Log("debug", "Trying to open $FILEF\n");
    if ($self->openfile($FILEF)) {
      $FILEF=$self->getParameter('MyFileHandle');
    }
    else {
      $FILEF=undef;
    }
  }
  else {
    $self->Log("debug", "Not opening any file.\n");
  }
}

sub openfile {
  my $self=checkref(shift);
  my $fname=shift;
  if ($fname) {
    my $file=IO::File->new($fname);
    if (!$file) {
      $self->Log("errors", "Error opening $fname: $!\n");
      return undef;
    }
    $self->Log("debug", "File $fname opened successfully.\n");
    $self->setParameter(MyFileHandle => $file);
    #$file->seek(0,2);
  }
  else {
    $self->Log("errors", "No file name provided, not doing anything.\n");
  }
}

=pod

Related to this, we have the C<getline> subroutine, which must return
a single line of text containing a record of information obtained from
the log sources. In its default form (and corresponding to the default
form of C<Init_log>), it simply reads a line from the open log file
and returns it.

Notice that the C<getline> subroutine must return C<undef> every once
in a while, otherwise the Filter will enter an inifinite cycle. This
is not a problem for the default case because eventually C<getline>
will reach the end of the file and will return C<undef> when there
is nothing else to read (for the moment), but it has to be taken
care of specifically when data is read from other sources.

=cut

sub getline {
  my $self=checkref(shift);
  # Read only if the file is open.
  if (defined($FILEF)){
    $self->Log("debug", "Trying to read from handle $FILEF.\n");
    return AAFID::Comm::nextLine(Handle => $FILEF, Timeout => 0);
  }
  else {
    $self->Log("debug", "No handle open, not reading anyting.\n");
    return undef;
  }
}

=pod

The connection between an Agent and a Filter is established by a
UNIX-domain socket which is a local socket to the current host.

=cut

sub newServerSocket {
  my $arch_sock=shift || $NAME;
  (-e $NAME) and do { unlink($NAME) or do { $!="Could not remove $NAME: $!";
					  return undef;
					}
		  }; 
  my $sock=IO::Socket::UNIX->new(Listen   => 10,
                                 Type           => SOCK_STREAM,
                                 Local          => $arch_sock,
				);
  # This is a hack so that AAFID::Comm::isServerSocket can detect
  # the type of socket automatically.
  $sock and $ {*$sock}{_AAFID_Comm_ServerSocket}=1;
  return $sock;
}

=pod

The initialization sets up the socket name where the filter will listen
for connections, and sets some global variables to make accessing
some of the parameters easier.

=cut

sub Init {
  my $self=checkref(shift);
  my $class=ref($self);
  $class=~ tr/A-Z/a-z/;
  my $route='/usr/tmp/aafid.'.$class;
  $self->setParameter( SocketPath => $route );
  $NAME=$self->getParameter('SocketPath');
  $FILEF=$self->getParameter('LogFile_Name');
  $PERIOD=$self->getParameter('CheckPeriod');

  # Initialize the source of logging info.
  $self->Init_log;

  # Filter establishes the UNIX-domain socket
  $Server=newServerSocket($NAME);
  if ($Server) {
    $self->Log("processes", "Filter started on $NAME\n");
    
    # Add server socket to the select object.
    $self->getParameter('OpenHandles')->add($Server);
    return $self;
  }
  else {
    $self->Log("errors", "Error starting filter server socket: $!\n");
    return undef;
  }
}

=head1 Getting the patterns and Selecting the information.

The agent sends patterns to the filter in order to select specific information.
The filter split the lines into fields to be more careful about what 
information is choosen. In this way, the filter looks for the line on the
log which matches with the pattern and reconstructs it to the original
format. Once this process is done, the filter send the information in the
 original format to the agent.

The agent uses the SETPATTERN command to tell the filter the pattern that
it is looking for. The arguments to this command must be the field names
of the data that the filter handles, with each one containing a regular
expression that is to be matched against that particular field. Only
log records that match on all the given fields are returned to the
agent.

For example, the following command:

   SETPATTERN Daemon => 'telnet', From => '.*\.cs\.purdue\.edu$ '

will cause the filter to return only records whose Daemon field contains
"telnet" B<and> whose From field ends with ".cs.purdue.edu".

=cut

sub command_SETPATTERN {
  my $self= shift;
  my $msg= shift;
  my %param=@_;
  my $result=$self->getParameter('OpenHandleIDs');
  $result->{$msg->{FROM}}->{Pattern}=\%param;
  $self->Log("debug", "Pattern for ".$msg->{FROM}." set to ".Dumper(\%param). "\n");
  return undef;
}

=pod

In order to give the filter some semantic knowledge about the data it
is processing, all the records are internally split in fields, and
processed as such. This allows the specification of per-field patterns
for matching, as is done by the SETPATTERN command. However, how to
split data into fields is completely dependent on the specific data
that is being processed, and thus is a functionality that needs to be
specified by the filter author. The C<makefield> subroutine takes as
argument a string, and returns a hash reference where each key is the
field name, and each element contains the corresponding field.

The filter author can provide her own C<makefield> subroutine that is
customized to the data that is being processed. However, for the most
low-denominator case, a default behavior is provided. The base
C<makefield> subroutine works as follows:

=over 4

=item *

If the C<DataFields> parameter is defined, it must contain an
array reference where each element corresponds to a field name, in the
order they appear in the string. The string is split into as many
fields as elements that array contains, considering the fields to be
separated by blank space, and the last field to "swallow" the rest of
the string.

=item *

If the C<DataFields> parameter is not defined, the string is
split into blank-space-separated fields, named "Field0", "Field1",
etc.

=back 4

=cut

sub makefield {
  my $self=checkref(shift);
  my $line=shift;
  my $fields=$self->getParameter('DataFields');
  if ($fields) {
    my @fields=@$fields;
    # Use the fields from the DataFields parameter.
    my $nfields=@fields;
    my @data=split(" ", $line, $nfields);
    my $data;
    my %res=map { $data=shift @data || "";
		  $_ => $data ;
		} @fields;
    return \%res;
  }
  else {
    # Use generic named fields.
    my $nfield=0;
    my $field;
    my %res=map { $field="Field" . $nfield++; 
		  $field => $_ 
		} split(" ", $line);
    return \%res;
  }
}

=pod

In order to send the data back to the agent, it needs to be
reconstructed from the field representation to a string
representation. This is what the C<makeline> subroutine does. It takes
a hash reference as an argument that contains the data fields, and
returns a string.

In its default form, C<makefield> concatenates (separated by spaces)
the fields in the order given by the C<DataFields> parameter, or
in numeric order if that parameter is not defined. Of course, the
filter author can override this subroutine to provide adequate
behavior if the default behavior is not correct.

If a second argument is given to the default form of C<makeline>, it is
used as a field separator instead of a space.

=cut

sub makeline {
  my $self=checkref(shift);
  my $data=shift;
  my $sep=shift || " ";
  my $fields=$self->getParameter('DataFields');
  my @fields;
  if ($fields) {
    # Join in order dictated by DataFields
    @fields=@$fields;
  }
  else {
    # Join in numeric order.
    @fields=sort keys %$data;
  }
  return join $sep, map { $data->{$_} || "" } @fields;
}

=pod

The C<run> subroutine is the heart of the filter. It checks for input
from any of its agents or from the log source, and does the right thing.

=cut

sub run {
  my $self=checkref(shift);
  my $i;
  my $socket;
  my $info_log;
  my $info_usr;
  my $field;
  my $flag;
  # Do some setup
  $self->setupDefaultSignalHandlers;
  # Send a CONNECT message
  $self->sendConnectMessage;
  while(1){
    $self->processInput($PERIOD);
    while ($_=$self->getline) {
      $self->Log("debug", "Got line: $_\n");
      $info_log=$self->makefield($_);
      # Skip if makefield tells us it's invalid.
      next unless $info_log;
      my $ohids=$self->getParameter('OpenHandleIDs');
      if (!$ohids) {
	croak "Panic: I don't have my OpenHandleIDs parameter!";
      }
      foreach $i (keys(%$ohids)) {
	$self->Log("debug", "Checking for $i\n");
   	$socket=$ohids->{$i}->{Handle};
        $flag=0;
   	if (defined($ohids->{$i}->{Pattern}) &&
	    %{$ohids->{$i}->{Pattern}}) {
	  $info_usr=$ohids->{$i}->{Pattern};
	  foreach $field (keys(%$info_usr)) {
	    if ($field =~ "All") {
	      print $socket $self->makeline($info_log)."\n";
	      $flag=1;
	    }
	    else {
	      if (defined($info_log->{$field})) {
		if ($info_log->{$field} =~ $info_usr->{$field}){
		  $flag+=0;
		}
		else {
		  $flag+=1;
		}       
	      }
	      else {
		$flag=1;
	      }
	    }
	  }
	  if (!$flag) {
	    # A match.
	    $self->Log("debug", "Got a match, sending line.\n");
	    print $socket $self->makeline($info_log)."\n";
	  }
	  else {
	    $self->Log("debug", "No match.\n");
	  }
        }
	else {
	  # We fall here is no pattern has been given, or an empty
	  # pattern has been given. So an empty pattern means
	  # "no matches", which is what we want.
	  $self->Log("debug", "Empty pattern, no match\n");
	}
      }
    }
  }
}

=pod

We use a modified C<sendConnectMessage> subroutine, which sends the
message with the subtype FILTER instead of CHILD, and the socket path
as the second element in the DATA field.

=cut

sub sendConnectMessage {
  my $self=checkref(shift);
  my $to=shift;
  my $data=$self->getParameter("ClassID") . " " .
    $self->getParameter("SocketPath") . " " .
      $self->getParameter("Description");
  my $msg=$self->newMsg(TYPE	=> "CONNECT",
			SUBTYPE	=> ($to?"PARENT":"FILTER"),
			DATA	=> $data);
  if ($to) {
    $self->Log("I/O", "Sending CONNECT message to $to: ".$msg->toString."\n");
    $self->sendMsgTo($msg, $to)
  }
  else {
    $self->Log("I/O", "Sending CONNECT message up: ".$msg->toString."\n");
    $self->sendReport($msg);
  }
  return $self;
}

_EndOfEntity;

#
# $Log: Filter.pm,v $
# Revision 1.3  1998/09/09 09:05:37  zamboni
# * Filter.pm: General debugging.
#
# Revision 1.2  1998/09/07 17:35:20  zamboni
# Added support for filters!
#
# * Monitor.pm: Replaced all uses of AAFID::Message->new by uses
#   of $self->newMsg.
#
# * Filter.pm: Cleaned up the code and the comments, and added some
#   extra features, such as default versions of makefield and
#   makeline that make it easier to define new filters if they read
#   data where the fields are space-separated. Also, added some
#   error checks here and there.
#
# * Entity.pm: Added filter support
#   - Added descriptions for FiltersNeeded and FilterPaths parameters.
#   - Modified getParameter so that if a second argument is given, it is
#     used to initialize the parameter in case it is undefined.
#   - Added subroutines connectFilters, newFilterClientSocket,
#     setFilterPattern
#   - Added subroutine newMsg.
#
# * ControllerEntity.pm: Added filter support.
#   - Some general cleanup (removing stray commented code, etc.)
#   - Rewrote some sections for clarity (in invoke, _invoke,
#     _instantiate_and_run)
#   - Modularized loadmodule. Created new subroutines:
#       _loadmodule_error
#       _loadmodule_success
#   - Modularized _instantiate_and_run, creating new subroutine
#       _fork_and_run.
#   - Added the setupFilters subroutine, which takes care of loading
#     all the filters required by an entity. It is complemented by
#     an augmented message_CONNECT, which detects when a filter is
#     connecting and does the necessary updating.
#
# * Comm.pm: Removed an annoying debug message from bufferInput.
#
# * Agent.pm: Added code to contact the filters.
#
# Revision 1.1  1998/09/02 15:58:29  zamboni
# Base class for filters.
#
